import asyncio
import concurrent.futures
import functools
from multiprocessing import Manager
from multiprocessing.managers import DictProxy
import os
import sys
import time
import traceback
from typing import Optional
from unittest.mock import patch

from lymbo import color
from lymbo.env import LYMBO_TEST_SCOPE_GLOBAL
from lymbo.item import TestItem
from lymbo.item import TestPlan
from lymbo.log import logger
from lymbo.log import trace_call
from lymbo.resource_manager import global_queue  # noqa: F401
from lymbo.resource_manager import manage_resources
from lymbo.resource_manager import prepare_scopes
from lymbo.resource_manager import set_scopes
from lymbo.resource_manager import unset_scope


@trace_call
def run_test_plan(test_plan: TestPlan, max_workers: Optional[int] = None) -> int:
    global global_queue

    # TODO add a try first to execute long test first
    # TODO shuffle the tests

    tstart = time.time()

    with Manager() as manager:

        scopes = prepare_scopes(test_plan, manager)

        run_tests_with_scopes = functools.partial(run_tests, scopes=scopes)
        manage_resources_with_scopes = functools.partial(
            manage_resources, scopes=scopes
        )

        if max_workers is None:
            max_workers = os.cpu_count()
            if max_workers is None:
                max_workers = 1

        logger().debug(f"run_test_plan - max_workers={max_workers}")

        with concurrent.futures.ProcessPoolExecutor(
            max_workers=max_workers
        ) as resources_manager:

            # # Start the resources manager processes
            resources_manager_futures = [
                resources_manager.submit(manage_resources_with_scopes)
                for _ in range(max_workers)
            ]

            with concurrent.futures.ProcessPoolExecutor(
                max_workers=max_workers
            ) as tests_executor:
                execresult = tests_executor.map(run_tests_with_scopes, test_plan)

            for r in execresult:
                pass  # TODO log result

            # should already be 0 but we force this value because this is what stop the resources manager processes
            with scopes[LYMBO_TEST_SCOPE_GLOBAL]["lock"]:
                scopes[LYMBO_TEST_SCOPE_GLOBAL]["count"] = 0

            for _ in range(max_workers if max_workers else 4):
                global_queue.put({"stop": True})

            # Wait for a maximum of 30 seconds for all resource managers to complete
            try:
                for future in concurrent.futures.as_completed(
                    resources_manager_futures, timeout=30
                ):
                    try:
                        _ = future.result()  # TODO log result
                    except concurrent.futures.TimeoutError:
                        logger().debug(
                            "run_test_plan - A resource manager process timed out while waiting for completion."
                        )
                    except Exception as e:
                        logger().debug(
                            f"run_test_plan - An error occurred while waiting the resource manager processes. exception=[{e}]"
                        )
            except concurrent.futures.TimeoutError:
                logger().debug(
                    "run_test_plan - Not all resource manager processes completed within 30 seconds."
                )

        # TODO ensure all the processes have been stopped and the resources released.

    duration = int(time.time() - tstart)

    return duration


def run_tests(tests: list[TestItem], scopes: DictProxy):
    """Run a group of tests sequentially."""

    try:
        set_scopes(scopes)

        for test_item in tests:

            try:
                with patch.dict(
                    os.environ,
                    test_item.scopes,
                ):
                    run_test(test_item)
            except Exception as ex:
                print("ERROR RUN_F " + str(ex) + "\n" + traceback.format_exc())
            unset_scope(scopes, test_item)

    except Exception as ex:
        print("ERROR RUN_TESTS " + str(ex))


@trace_call
def run_test(test_item: TestItem):

    path = test_item.path
    asynchronous = test_item.asynchronous
    name = test_item.fnc
    cls = test_item.cls
    args, kwargs = test_item.parameters

    print(".", end="", flush=True)

    syspath = sys.path + [
        str(path.parent.absolute()),
    ]

    with patch.object(sys, "path", syspath):

        module = __import__(str(path.name[:-3]))
        if cls:
            classdef = getattr(module, cls)
            self = classdef()
            test_function = getattr(self, name)
        else:
            test_function = getattr(module, name)

        try:
            test_item.start()
            if asynchronous:
                asyncio.run(test_function(*args, **kwargs))
            else:
                test_function(*args, **kwargs)
            test_item.end()
            print(f"{color.GREEN}P{color.RESET}", end="", flush=True)
        except AssertionError as ex:
            test_item.end(reason=ex)
            print(f"{color.RED}F{color.RESET}", end="", flush=True)
        except Exception as ex:
            test_item.end(reason=ex)
            print(f"{color.YELLOW}B{color.RESET}", end="", flush=True)



from enum import Enum
from functools import cache
import logging
import os
from pathlib import Path
from typing import Union

from lymbo.env import LYMBO_LOG_LEVEL
from lymbo.env import LYMBO_LOG_PATH


class LogLevel(Enum):
    NOTSET = "notset"
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"


def set_env_for_logging(log_level: LogLevel, path: Union[Path, None]):

    os.environ[LYMBO_LOG_LEVEL] = str(
        {
            "NOTSET": logging.NOTSET,
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
            "CRITICAL": logging.CRITICAL,
        }[log_level.value.upper()]
    )

    if path:
        os.environ[LYMBO_LOG_PATH] = str(path)


@cache
def logger() -> logging.Logger:

    logger = logging.getLogger()
    log_level = int(os.environ.get(LYMBO_LOG_LEVEL, logging.WARNING))
    logger.setLevel(log_level)

    formatter = logging.Formatter(
        "¤ %(asctime)s.%(msecs)03d (%(process)d) [%(levelname)s] %(message)s",
        datefmt="%H:%M:%S",
    )

    if LYMBO_LOG_PATH in os.environ:
        file_handler = logging.FileHandler(os.environ[LYMBO_LOG_PATH])
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    return logger


def trace_call(fnc):

    def wrapper(*args, **kwargs):

        log_level = int(os.environ.get(LYMBO_LOG_LEVEL, logging.WARNING))

        if log_level == logging.DEBUG:
            try:
                logger().debug(f"{fnc.__name__} | begin args={args} kwargs={kwargs}")
            except Exception as ex:
                logger().debug(f"trace_call - exception - [{ex}]")

        r = fnc(*args, **kwargs)

        if log_level == logging.DEBUG:
            try:
                logger().debug(f"{fnc.__name__} |end r={r}")
            except Exception as ex:
                logger().debug(f"trace_call - exception - [{ex}]")

        return r

    return wrapper



import contextlib
import os

from lymbo.env import LYMBO_TEST_COLLECTION


@contextlib.contextmanager
def test(args=None):
    yield


def args(*args, **kwargs):
    """
    Define the parameters used to call the test function.

    If a parameter is defined using the `lymbo.expand` function,
    the resulting list is flattened before being passed to the test function.
    """

    if LYMBO_TEST_COLLECTION in os.environ:

        flattened_params = [(args, kwargs)]

        pos = 0
        for arg in args:
            if type(arg) is ArgParams:
                new_flattened_params = []
                for elt in arg.args:
                    for params in flattened_params:
                        gargs, gkwargs = params
                        gargs = list(gargs)
                        gargs[pos] = elt
                        gargs = tuple(gargs)
                        new_flattened_params.append((gargs, gkwargs))
                flattened_params = new_flattened_params
            pos += 1

        for key, value in kwargs.items():
            if type(value) is ArgParams:
                new_flattened_params = []
                for elt in value.args:
                    for params in flattened_params:
                        gargs, gkwargs = params
                        gkwargs = gkwargs.copy()
                        gkwargs[key] = elt
                        new_flattened_params.append((gargs, gkwargs))
                flattened_params = new_flattened_params

        return flattened_params


class ArgParams:

    def __init__(self, *args):
        self.args = list(*args)


def expand(*args) -> ArgParams:
    return ArgParams(args)



import glob
import os
import tempfile

from lymbo.env import LYMBO_REPORT_PATH


class TestReport:

    def __init__(self, path=None):
        if path is None:
            path = tempfile.mkdtemp()
        else:
            os.makedirs(path, exist_ok=True)
        self.path = path

        os.environ[LYMBO_REPORT_PATH] = str(self.path)

        self.clean()

    def clean(self):

        for lymbofile in glob.glob(os.path.join(self.path, "lymbo-*")):
            os.remove(lymbofile)



import os
from pathlib import Path
import unittest

from lymbo.collect import collect_tests
from lymbo.item import GroupBy

dir = os.path.dirname(os.path.abspath(__file__))


class TestCollect(unittest.TestCase):

    # group by

    def test_collect_tests_groupby_none(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.NONE
        )

        self.assertEqual(test_plan.count, (11, 11))

    def test_collect_tests_groupby_function(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.FUNCTION
        )

        self.assertEqual(test_plan.count, (11, 6))

    def test_collect_tests_groupby_class(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.CLASS
        )

        self.assertEqual(test_plan.count, (11, 10))

    def test_collect_tests_groupby_module(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.MODULE
        )

        self.assertEqual(test_plan.count, (11, 2))

    # list files

    def test_collect_tests_dir_not_exists(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect_xxxxx"))], GroupBy.NONE
        )

        self.assertEqual(test_plan.count, (0, 0))

    def test_collect_tests_dir(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.NONE
        )

        self.assertEqual(test_plan.count, (11, 11))

    def test_collect_tests_one_file(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect/collect_a.py"))], GroupBy.NONE
        )

        self.assertEqual(test_plan.count, (4, 4))

    def test_collect_tests_two_files(self):

        test_plan = collect_tests(
            [
                Path(os.path.join(dir, "data_collect/collect_a.py")),
                Path(os.path.join(dir, "data_collect/collect_b.py")),
            ],
            GroupBy.NONE,
        )

        self.assertEqual(test_plan.count, (11, 11))


if __name__ == "__main__":
    unittest.main()



import argparse
from pathlib import Path

from lymbo.item import GroupBy
from lymbo.item import ReportFailure
from lymbo.log import LogLevel


def parse_args() -> argparse.Namespace:

    parser = argparse.ArgumentParser(description="lymbo")

    parser.add_argument(
        "paths",
        metavar="PATH",
        type=Path,
        nargs="*",
        default=[Path("test", Path("tests"))],
        help="Path(s) for test collection",
    )

    parser.add_argument(
        "--version", action="store_true", help="Print the version and exit."
    )

    parser.add_argument(
        "--collect", action="store_true", help="Print the test plan and exit."
    )

    parser.add_argument(
        "--groupby",
        type=GroupBy,
        choices=GroupBy,
        default=GroupBy.NONE,
        help="Grouped tests are executed sequentialy.",
    )

    parser.add_argument(
        "--report", type=Path, help="Save the report in that directory."
    )

    parser.add_argument(
        "--log-level",
        type=LogLevel,
        choices=LogLevel,
        default=LogLevel.WARNING,
        help="The log level",
    )

    parser.add_argument("--log", type=Path, help="Path to the log file.")

    parser.add_argument(
        "--report-failure",
        type=ReportFailure,
        choices=ReportFailure,
        default=ReportFailure.NORMAL,
        help="The log level",
    )

    parser.add_argument(
        "--workers",
        type=int,
        default=None,
        help="The number of workers in parrallel (default = number of CPU)",
    )

    return parser.parse_args()



import contextlib
import importlib
import inspect
import io
from multiprocessing import Queue
from multiprocessing.managers import DictProxy
from multiprocessing.managers import SyncManager
import os
from pathlib import Path
import pickle
import sys
import time
import traceback
from typing import Union
from unittest.mock import patch

from lymbo.env import LYMBO_TEST_SCOPE_CLASS
from lymbo.env import LYMBO_TEST_SCOPE_FUNCTION
from lymbo.env import LYMBO_TEST_SCOPE_GLOBAL
from lymbo.env import LYMBO_TEST_SCOPE_MAX
from lymbo.env import LYMBO_TEST_SCOPE_MODULE
from lymbo.exception import LymboExceptionScopeHierarchy
from lymbo.exception import LymboExceptionScopeNested
from lymbo.item import TestItem
from lymbo.item import TestPlan
from lymbo.log import logger


shared_scopes: Union[DictProxy, None] = None
global_queue: Queue = Queue()


def set_scopes(scopes):
    global shared_scopes
    shared_scopes = scopes


@contextlib.contextmanager
def _cm_by_scope(scope_name, cm, *args, **kwargs):

    global shared_scopes
    scopes = shared_scopes
    global global_queue

    unique_cm_id = f"{cm.__module__}.{cm.__name__}.{args}.{kwargs}"

    scope = scopes[os.environ[scope_name]]
    with scope["lock"]:  # the lock is only for the request about the resource creation

        if unique_cm_id not in scope["resources"]:
            scope["resources"][unique_cm_id] = None
            scope["resources_output"][unique_cm_id] = ""

            module_name = cm.__module__
            module = importlib.import_module(module_name)
            module_path = inspect.getfile(module)

            global_queue.put(
                {
                    "stop": False,
                    "scope_id": os.environ.get(scope_name),
                    "resource": {
                        "id": unique_cm_id,
                        "module_name": module_name,
                        "module_path": module_path,
                        "name": cm.__name__,
                        "args": args,
                        "kwargs": kwargs,
                        "environ": os.environ.copy(),
                    },
                }
            )

        while scope["resources"][unique_cm_id] is None:
            time.sleep(0.1)  # TODO infinite loop risk

        if scope["resources_output"][unique_cm_id]:
            print(
                scope["resources_output"][unique_cm_id]
            )  # by printing the output here, it will be added to the test output

        resource = pickle.loads(scope["resources"][unique_cm_id])

    if isinstance(resource, Exception):
        raise resource  # TODO report the original traceback

    yield resource


@contextlib.contextmanager
def scope_global(cm, *args, **kwargs):

    if LYMBO_TEST_SCOPE_MAX not in os.environ:
        raise LymboExceptionScopeNested(
            "You can't share a resource in another shared resource."
        )

    if os.environ[LYMBO_TEST_SCOPE_MAX] in ("module", "class", "function"):
        raise LymboExceptionScopeHierarchy(
            f"You can't share a resource with the scope [global] under a shared resource with the scope [{os.environ[LYMBO_TEST_SCOPE_MAX]}]"
        )

    with patch.dict(os.environ, {LYMBO_TEST_SCOPE_MAX: "global"}):

        with _cm_by_scope(LYMBO_TEST_SCOPE_GLOBAL, cm, *args, **kwargs) as resource:
            yield resource


@contextlib.contextmanager
def scope_module(cm, *args, **kwargs):

    if LYMBO_TEST_SCOPE_MAX not in os.environ:
        raise LymboExceptionScopeNested(
            "You can't share a resource in another shared resource."
        )

    if os.environ[LYMBO_TEST_SCOPE_MAX] in ("class", "function"):
        raise LymboExceptionScopeHierarchy(
            f"You can't share a resource with the scope [module] under a shared resource with the scope [{os.environ[LYMBO_TEST_SCOPE_MAX]}]"
        )

    with patch.dict(os.environ, {LYMBO_TEST_SCOPE_MAX: "module"}):

        with _cm_by_scope(LYMBO_TEST_SCOPE_MODULE, cm, *args, **kwargs) as resource:
            yield resource


@contextlib.contextmanager
def scope_class(cm, *args, **kwargs):

    if LYMBO_TEST_SCOPE_MAX not in os.environ:
        raise LymboExceptionScopeNested(
            "You can't share a resource in another shared resource."
        )

    if os.environ[LYMBO_TEST_SCOPE_MAX] == "function":
        raise LymboExceptionScopeHierarchy(
            f"You can't share a resource with the scope [class] under a shared resource with the scope [{os.environ[LYMBO_TEST_SCOPE_MAX]}]"
        )

    with patch.dict(os.environ, {LYMBO_TEST_SCOPE_MAX: "class"}):

        with _cm_by_scope(LYMBO_TEST_SCOPE_CLASS, cm, *args, **kwargs) as resource:
            yield resource


@contextlib.contextmanager
def scope_function(cm, *args, **kwargs):

    if LYMBO_TEST_SCOPE_MAX not in os.environ:
        raise LymboExceptionScopeNested(
            "You can't share a resource in another shared resource."
        )

    with patch.dict(os.environ, {LYMBO_TEST_SCOPE_MAX: "function"}):

        with _cm_by_scope(LYMBO_TEST_SCOPE_FUNCTION, cm, *args, **kwargs) as resource:
            yield resource


def new_scope(manager: SyncManager) -> DictProxy:
    """Define a scope and its resources in a variable shared among all processes."""
    scope = manager.dict()
    scope["count"] = 0  # number of possible occurrence for this scope
    scope["lock"] = manager.Lock()  # to update the "count" value
    scope["resources"] = manager.dict()  # all the resources created for this scope
    scope["resources_output"] = (
        manager.dict()
    )  # all the resources output created for this scope
    return scope


def prepare_scopes(test_plan: TestPlan, manager: SyncManager) -> DictProxy:
    """Prepare a list of all possible scopes and store them in a variable shared among all processes."""

    scopes = manager.dict()

    # The scopes are retrieved from the test plan.

    for tests in test_plan:
        for test in tests:
            for scope_category, scope in test.scopes.items():

                if scope not in scopes:
                    scopes[scope] = new_scope(manager)

                scopes[scope]["count"] += 1

    return scopes


def manage_resources(scopes):

    global global_queue

    resources = {}

    while scopes[LYMBO_TEST_SCOPE_GLOBAL]["count"] > 0:

        message = global_queue.get()

        if message["stop"]:
            break  # all the tests have been executed

        # setup resource
        try:
            module_name = message["resource"]["module_name"]
            module_path = Path(message["resource"]["module_path"])
            name = message["resource"]["name"]
            args = message["resource"]["args"]
            kwargs = message["resource"]["kwargs"]
            environ = message["resource"]["environ"]
            scope_id = message["scope_id"]
            scope = scopes[scope_id]

            syspath = sys.path + [
                str(module_path.parent.absolute()),
            ]

            if LYMBO_TEST_SCOPE_MAX in environ:
                del environ[
                    LYMBO_TEST_SCOPE_MAX
                ]  # to detect if we try to create a shared resource from another shared resource

            with (
                patch.object(sys, "path", syspath),
                patch.dict(os.environ, environ),
            ):

                original_stdout = sys.stdout
                original_stderr = sys.stderr

                stdout = io.StringIO()
                sys.stdout = stdout
                sys.stderr = stdout

                module = importlib.import_module(module_name)

                ctxmgr = getattr(module, name)

                cm = ctxmgr(*args, **kwargs)

                try:
                    logger().debug(
                        f"manage_resources - instantiate a resource for {scope_id} -> "
                        f"resource=[{module_name}.{name}({args}{kwargs})]"
                    )
                    resource = cm.__enter__()
                    logger().debug(
                        f"manage_resources - instantiate a resource for {scope_id} -> "
                        f"resource=[{module_name}.{name}({args}{kwargs}] done"
                    )
                except Exception as ex:
                    resource = ex

                scope["resources"][message["resource"]["id"]] = pickle.dumps(resource)

                sys.stdout = original_stdout
                sys.stderr = original_stderr

                scope["resources_output"][message["resource"]["id"]] = stdout.getvalue()

                # we save the context manager to execute the teardown method when the scope count =0
                resources[scope_id] = resources.get(scope_id, [])
                resources[scope_id].append(cm)

        except Exception as ex:
            logger().debug(
                f"manage_resources - exception during the instatiation of a resource for {scope_id} -> "
                f"resource=[{module_name}.{name}({args}{kwargs}] Exception=[{ex}]"
            )

        # free resources
        teardown_resources(scopes, resources)

    # free resources
    teardown_resources(scopes, resources)


def teardown_resources(scopes, resources):

    released_scopes = []

    try:

        for scope_id, resources_by_scope in resources.items():
            if scopes[scope_id]["count"] == 0:
                for resource in resources_by_scope:
                    try:
                        logger().debug(
                            f"teardown_resources - teardown resource for {scope_id} -> resource=[{resource}]"
                        )
                        original_stdout = sys.stdout
                        original_stderr = sys.stderr

                        stdout = io.StringIO()
                        sys.stdout = stdout
                        sys.stderr = stdout

                        resource.__exit__(
                            None, None, None
                        )  # TODO pass exception if necessary

                        sys.stdout = original_stdout
                        sys.stderr = original_stderr
                    except Exception as ex:
                        sys.stdout = original_stdout
                        sys.stderr = original_stderr

                        logger().warning(
                            "An exception occurred during the execution of the resource's teardown."
                            f" stdout=[{stdout.getvalue()}], exception=[{ex}], traceback={traceback.format_exc()}"
                        )
                released_scopes.append(scope_id)

        released_scopes = set(released_scopes)

        for scope_id in released_scopes:
            del resources[scope_id]
    except Exception as ex:
        logger().debug(
            f"teardown_resources exception=[{ex}], traceback={traceback.format_exc()}"
        )


def unset_scope(scopes, test_item: TestItem):

    for scope_id in test_item.scopes.values():
        scope = scopes[scope_id]
        with scope["lock"]:
            scope["count"] -= 1



import json
import os
from pathlib import Path
import unittest

from lymbo.collect import collect_tests
from lymbo.item import GroupBy
from lymbo.run import run_test_plan
from lymbo.report import TestReport

dir = os.path.dirname(os.path.abspath(__file__))


class TestResource(unittest.TestCase):

    def assert_same_resource(self, resources, scope, nb):
        self.assertEqual(len(resources[scope]), nb, resources[scope])
        self.assertEqual(len(set(resources[scope])), 1, resources[scope])

    def test_resource(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_resource/"))], GroupBy.NONE
        )

        _ = TestReport()

        run_test_plan(test_plan)

        resources = {}

        for group in test_plan:
            for test in group:
                test.refresh_from_report()
                resource = json.loads(test.output.getvalue())
                resources[resource["scope"]] = resources.get(resource["scope"], [])
                resources[resource["scope"]].append(resource["value"])

        with self.subTest("use context manager without scope"):
            # all the resources are unique
            self.assertEqual(len(set(resources["none"])), 4, resources["none"])

        with self.subTest("use context manager with scope function"):
            # all the resources under the same scope are identical
            self.assert_same_resource(resources, "function_a_1", 4)
            self.assert_same_resource(resources, "function_a_2", 3)
            self.assert_same_resource(resources, "function_b_1", 3)
            self.assert_same_resource(resources, "function_b_2", 1)

        with self.subTest("use context manager with scope class"):
            # all the resources under the same scope are identical
            self.assert_same_resource(resources, "class_a_1", 2)
            self.assert_same_resource(resources, "class_a_2", 2)
            self.assert_same_resource(resources, "class_b_1", 2)
            self.assert_same_resource(resources, "class_b_2", 4)

        with self.subTest("use context manager with scope module"):
            # all the resources under the same scope are identical
            self.assert_same_resource(resources, "module_a", 6)
            self.assert_same_resource(resources, "module_b", 3)

        with self.subTest("use context manager with scope global"):
            # all the resources under the same scope are identical
            self.assert_same_resource(resources, "global", 10)


if __name__ == "__main__":
    unittest.main()



import json
import os
from pathlib import Path
import unittest

from lymbo.collect import collect_tests
from lymbo.item import GroupBy
from lymbo.item import TestStatus
from lymbo.run import run_test_plan
from lymbo.report import TestReport

dir = os.path.dirname(os.path.abspath(__file__))


class TestRunTestPlan(unittest.TestCase):

    # max_workers

    def test_run_max_workers_default(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_run/run_a.py"))], GroupBy.NONE
        )

        _ = TestReport()

        run_test_plan(test_plan)

        workers_pids = []

        for group in test_plan:
            for test in group:
                test.refresh_from_report()
                trace = json.loads(test.output.getvalue())
                workers_pids.append(trace["pid"])

        self.assertEqual(len(set(workers_pids)), os.cpu_count())

    def test_run_max_workers_custom(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_run/run_a.py"))], GroupBy.NONE
        )

        _ = TestReport()

        run_test_plan(test_plan, 2)

        workers_pids = []

        for group in test_plan:
            for test in group:
                test.refresh_from_report()
                trace = json.loads(test.output.getvalue())
                workers_pids.append(trace["pid"])

        self.assertEqual(len(set(workers_pids)), 2)


class TestTestItemStatus(unittest.TestCase):

    def test_status(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_run/run_status.py"))], GroupBy.MODULE
        )

        _ = TestReport()

        run_test_plan(test_plan, 1)

        status = {}

        for group in test_plan:
            for test in group:
                test.refresh_from_report()
                status[test.fnc] = test.status

        with self.subTest("the test passed"):
            self.assertEqual(status["passed"], TestStatus.PASSED)

        with self.subTest("the test is broken"):
            self.assertEqual(status["broken"], TestStatus.BROKEN)

        with self.subTest("the test failed"):
            self.assertEqual(status["failed"], TestStatus.FAILED)


if __name__ == "__main__":
    unittest.main()



import platform
import sys

import lymbo
from lymbo.collect import collect_tests
from lymbo.config import parse_args
from lymbo.item import TestStatus
from lymbo.log import set_env_for_logging
from lymbo.report import TestReport
from lymbo.run import run_test_plan


def lymbo_entry_point():

    config = parse_args()

    set_env_for_logging(config.log_level, config.log)

    if config.version:
        print(lymbo.__version__)
        sys.exit(5)

    print(
        f"** lymbo {lymbo.__version__} (python {platform.python_version()}) ({platform.platform()}) **"
    )

    print("==== collecting tests")
    test_plan = collect_tests(config.paths, config.groupby)

    if config.collect:
        test_plan_to_print, _ = test_plan.test_plan(show_status=False)
        print(test_plan_to_print)

    nb_tests, nb_groups = test_plan.count
    print(
        f"==== {nb_tests} test{'s' if nb_tests>1 else ''} in {nb_groups} group{'s' if nb_groups>1 else ''}"
    )

    if config.collect:
        sys.exit(0)

    if nb_tests == 0:
        sys.exit(5)

    _ = TestReport(config.report)

    print("==== running tests")

    duration = run_test_plan(test_plan, config.workers)

    print(f"\n==== tests executed in {duration} second{'s' if duration>1 else ''}")

    print("==== results")
    test_plan_to_print, tests_status = test_plan.test_plan(show_status=True)
    print(test_plan_to_print)

    print(
        f"==== {''.join([(f'{nb} {status.value} ') for status, nb in tests_status.items() if nb > 0])} "
    )

    if (tests_status[TestStatus.BROKEN] > 0) or (tests_status[TestStatus.FAILED] > 0):
        print("==== failures")

        failures, nb_failures = test_plan.failures(report_failure=config.report_failure)
        print(failures)

        print("====")

        if nb_failures > 0:
            exit(1)


if __name__ == "__main__":
    lymbo_entry_point()
    exit(0)
